# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.views.decorators.csrf import csrf_exempt
from blti.config import get_tool_conf, get_launch_data_storage
from pylti1p3.contrib.django import DjangoOIDCLogin


def get_launch_url(request):
    try:
        return request.POST.get(
            'target_link_uri', request.GET.get('target_link_uri'))
    except KeyError:
        raise BLTIException('Missing "target_link_uri" param')


@csrf_exempt
def login(request):
    try:
        tool_conf = get_tool_conf()
        launch_data_storage = get_launch_data_storage()
        oidc_login = DjangoOIDCLogin(
            request, tool_conf, launch_data_storage=launch_data_storage)
        target_link_uri = get_launch_url(request)

        if target_link_uri.startswith('http:') and request.is_secure():
            target_link_uri = f"https:{target_link_uri[5:]}"

        return oidc_login.enable_check_cookies().redirect(target_link_uri)
    except Exception as ex:
        return HttpResponse(str(ex), status=401)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


LTI_1p3_ROLES_CLAIM = "https://purl.imsglobal.org/spec/lti/claim/roles"
LTI_1p3_ROLE_PREFIX = "http://purl.imsglobal.org/vocab/lis/v2/"


def roles_from_role_name(role_names):
    roles = []

    for role_name in role_names:
        if role_name == 'User':
            roles += ["system/person#User"]
        elif role_name in ['Observer', 'Mentor']:
            roles += [
                "institution/person#Observer",
                "institution/person#Mentor",
                "membership#Mentor"]
        elif role_name in ['Student', 'Learner']:
            roles += [
                "institution/person#Learner",
                "institution/person#Student",
                "membeship#Learner"]
        elif role_name in ['Instructor', 'Faculty', 'Teacher']:
            roles += [
                "institution/person#Instructor",
                "institution/person#Faculty",
                "membeship#Instructor"]
        elif role_name == 'Administrator':
            roles += [
                "system/person#User",
                "institution/person#Administrator",
                "membeship#Administrator"]
        elif role_name == 'TeachingAssistant':
            roles += ["membership/Instructor#TeachingAssistant"]
        elif role_name == 'ContentDeveloper':
            roles += ["membership#ContentDeveloper"]

    return LTI_1p3_ROLES_CLAIM, [f"{LTI_1p3_ROLE_PREFIX}{r}" for r in roles]



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.http import HttpResponse
from .base import BLTIView
from blti.exceptions import BLTIException
from blti.performance import log_response_time
import json


class RESTDispatch(BLTIView):
    """
    A superclass for API views
    """
    authorized_role = 'member'

    @log_response_time
    def dispatch(self, request, *args, **kwargs):
        try:
            self.validate(request)
        except BLTIException as ex:
            return self.error_response(401, ex)

        return super(BLTIView, self).dispatch(request, *args, **kwargs)

    def render_to_response(self, context, **kwargs):
        return self.json_response(content=context)

    def error_response(self, status, message='', content={}):
        content['error'] = str(message)
        return self.json_response(content=content, status=status)

    def json_response(self, content={}, status=200):
        response = HttpResponse(json.dumps(content),
                                status=status,
                                content_type='application/json')
        self.add_headers(response=response)
        return response



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.test import RequestFactory, TestCase, override_settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import ImproperlyConfigured
from blti.validators import BLTIRequestValidator, Roles
from blti.models import CanvasData
from blti.roles import roles_from_role_name
from blti.performance import log_response_time
from blti import BLTI, LTI_DATA_KEY
from blti.mock_data import Mock1p3Data
from blti.exceptions import BLTIException
from oauthlib.common import generate_timestamp, generate_nonce
from urllib.parse import urlencode
import time
import mock
import os


class LTICanvasDataTest(TestCase):
    def test_known_product_code(self):
        # unimplemented product code
        data = Mock1p3Data().launch_data()
        data["https://purl.imsglobal.org/spec/lti/claim/tool_platform"][
            "product_family_code"] = "my_lms"
        self.assertRaises(ValueError, CanvasData, **data)

    def test_attributes(self):
        data = Mock1p3Data().launch_data()
        blti = CanvasData(**data)

        self.assertEquals(blti.link_title,
                          'PSYCH 101 A Au 19, Introduction To Psychology')
        self.assertEquals(blti.return_url,
                          ("https://uw.test.instructure.com/courses/"
                           "9752574/external_content/success/"
                           "external_tool_redirect"))
        self.assertEquals(blti.canvas_course_id, '9752574')
        self.assertEquals(blti.course_sis_id, '2019-autumn-PSYCH-101-A')
        self.assertEquals(blti.course_short_name, 'PSYCH 101 A')
        self.assertEquals(blti.course_long_name,
                          'PSYCH 101 A Au 19, Introduction To Psychology')
        self.assertEquals(blti.canvas_user_id, '700007')
        self.assertEquals(blti.user_login_id, 'javerage')
        self.assertEquals(blti.user_sis_id, '0C8F043FA5CBE23F2B1E1A63B1BD80B8')
        self.assertEquals(blti.user_full_name, 'James Average')
        self.assertEquals(blti.user_first_name, 'James')
        self.assertEquals(blti.user_last_name, 'Average')
        self.assertEquals(blti.user_email, 'javerage@u.washington.edu')
        self.assertEquals(
            blti.user_avatar_url, (
                "/images/thumbnails/8140331/"
                "ynu8th19hg8afjfy1y1bnzvi4nfewe7l6tsqf7kp"))
        self.assertEquals(blti.canvas_account_id, '5171292')
        self.assertEquals(blti.account_sis_id,
                          "uwcourse:seattle:a-and-s:pych:psych")
        self.assertEquals(blti.canvas_api_domain,
                          "uw.test.instructure.com")


class CanvasRolesTest(TestCase):
    def setUp(self):
        self.params = Mock1p3Data().launch_data()
        self.launch_data = CanvasData(**self.params)

    def _authorize(self, role):
        return Roles(self.launch_data).authorize(role=role)

    def test_authorize_nodata(self):
        # no blti object
        self.assertRaises(
            ImproperlyConfigured, Roles(None).authorize, role='member')

    def test_authorize_public(self):
        self.assertEquals(None, self._authorize('public'))
        self.assertEquals(None, self._authorize(None))

    def test_authorize_member(self):
        self._set_role('User')
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='member')

        self._set_role('Administrator')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('member'))

        self._set_role('Observer')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('member'))

    def test_authorize_admin(self):
        self._set_role('User')
        self.assertRaises(
            BLTIException, Roles(self.launch_data).authorize, role='admin')

        self._set_role('Learner')
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='admin')

        self._set_role('Administrator')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        self._set_role('TeachingAssistant')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        self._set_role('ContentDeveloper')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        self._set_role('Student')
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='admin')

    def test_authorize_specific(self):
        self._set_role('Learner')
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('Learner'))

        self._set_role('ContentDeveloper')
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='Instructor')

        # unknown role
        self._set_role('Learner')
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='Manager')

    def _set_role(self, role):
        claim, roles = roles_from_role_name([role])
        self.params[claim] = roles


class BLTI1p3SessionTest(TestCase):
    def setUp(self):
        self.request = RequestFactory().post(
            '/test', data=Mock1p3Data().launch_data(),
            secure=True)
        SessionMiddleware(get_response=mock.MagicMock()).process_request(
            self.request)

    def test_set_session(self):
        blti = BLTI()
        self.assertFalse(LTI_DATA_KEY in self.request.session)
        blti.set_session(self.request)
        self.assertTrue(LTI_DATA_KEY in self.request.session)

    def test_get_session(self):
        blti = BLTI()
        self.assertRaises(BLTIException, blti.get_session, self.request)

        blti.set_session(self.request)
        self.assertEqual(blti.get_session(self.request), {})



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.test import RequestFactory, TestCase, override_settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import ImproperlyConfigured
from blti.models import CanvasData
from blti.validators import BLTIRequestValidator, Roles
from blti.performance import log_response_time
from blti import BLTI, LTI_DATA_KEY
from blti.mock_data import Mock1p3Data
from blti.exceptions import BLTIException
from oauthlib.common import generate_timestamp, generate_nonce
from urllib.parse import urlencode
import time
import mock
import os


LTI_LAUNCH_PARAMS = {
    'oauth_consumer_key': 'XXXXXXXXXXXXXX',
    'oauth_signature_method': 'HMAC-SHA1',
    'oauth_timestamp': '1234567890',
    'oauth_nonce': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
    'oauth_signature': 'XXXXXXXXXXXXXXXXXXXXXXXXXXX=',
    'oauth_callback': 'about:blank',
    'oauth_version': '1.0',
    'launch_presentation_height': '400',
    'user_image': ('https://example.instructure.com/images/thumbnails/'
                   '123456/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'),
    'context_id': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
    'tool_consumer_info_version': 'cloud',
    'ext_roles': ('urn:lti:instrole:ims/lis/Administrator,'
                  'urn:lti:instrole:ims/lis/Instructor,'
                  'urn:lti:instrole:ims/lis/Student,'
                  'urn:lti:role:ims/lis/Instructor,'
                  'urn:lti:role:ims/lis/Learner/NonCreditLearner,'
                  'urn:lti:role:ims/lis/Mentor,urn:lti:sysrole:ims/lis/User'),
    'tool_consumer_instance_guid': 'xxxxxxxx.example.instructure.com',
    'context_label': 'ABC 101 A',
    'lti_message_type': 'basic-lti-launch-request',
    'custom_canvas_workflow_state': 'claimed',
    'lis_person_name_full': 'James Average',
    'context_title': 'ABC 101 A: Example Course',
    'user_id': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
    'custom_canvas_user_id': '123456',
    'launch_presentation_locale': 'en',
    'custom_canvas_api_domain': 'example.instructure.com',
    'custom_canvas_enrollment_state': 'active',
    'tool_consumer_instance_contact_email': 'example@example.instructure.com',
    'tool_consumer_info_product_family_code': 'canvas',
    'custom_application_type': 'ExampleApp',
    'lis_person_name_family': 'Average',
    'lis_course_offering_sourcedid': '2018-spring-ABC-101-A',
    'launch_presentation_width': '800',
    'resource_link_title': 'Example App',
    'custom_canvas_account_sis_id': 'example:account',
    'lis_person_sourcedid': 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
    'tool_consumer_instance_name': 'Example University',
    'resource_link_id': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
    'lis_person_contact_email_primary': 'javerage@example.edu',
    'roles': 'Learner',
    'custom_canvas_course_id': '123456',
    'lti_version': 'LTI-1p0',
    'lis_person_name_given': 'James',
    'launch_presentation_return_url': ('https://example.instructure.com/'
                                       'courses/123456'),
    'launch_presentation_document_target': 'iframe',
    'custom_canvas_account_id': '12345',
    'custom_canvas_user_login_id': 'javerage'
}


class RequestValidator1p1Test(TestCase):
    def setUp(self):
        self.request = RequestFactory().post(
            '/test', data=LTI_LAUNCH_PARAMS, secure=True)

    def test_check_client_key(self):
        self.assertTrue(BLTIRequestValidator().check_client_key('x' * 12))
        self.assertTrue(BLTIRequestValidator().check_client_key('5' * 30))
        self.assertTrue(BLTIRequestValidator().check_client_key('-' * 20))
        self.assertTrue(BLTIRequestValidator().check_client_key('_' * 20))
        self.assertFalse(BLTIRequestValidator().check_client_key('x' * 11))
        self.assertFalse(BLTIRequestValidator().check_client_key('x' * 31))
        self.assertFalse(BLTIRequestValidator().check_client_key('*' * 40))

    def test_check_nonce(self):
        self.assertTrue(BLTIRequestValidator().check_nonce('x' * 20))
        self.assertTrue(BLTIRequestValidator().check_nonce('5' * 50))
        self.assertTrue(BLTIRequestValidator().check_nonce('-' * 20))
        self.assertTrue(BLTIRequestValidator().check_nonce('_' * 20))
        self.assertFalse(BLTIRequestValidator().check_nonce('*' * 20))
        self.assertFalse(BLTIRequestValidator().check_nonce('x' * 19))
        self.assertFalse(BLTIRequestValidator().check_nonce('x' * 51))

    def test_validate_client_key(self):
        with self.settings(LTI_CONSUMERS={}):
            self.assertFalse(
                BLTIRequestValidator().validate_client_key('X', self.request))

        with self.settings(LTI_CONSUMERS={'A': '12345'}):
            self.assertTrue(
                BLTIRequestValidator().validate_client_key('A', self.request))

    def test_get_client_secret(self):
        with self.settings(LTI_CONSUMERS={}):
            self.assertEquals(
                BLTIRequestValidator().get_client_secret('X', self.request),
                'dummy')

        with self.settings(LTI_CONSUMERS={'A': '12345'}):
            self.assertEquals(
                BLTIRequestValidator().get_client_secret('A', self.request),
                '12345')

    def test_validate_timestamp_and_nonce(self):
        self.assertTrue(
            BLTIRequestValidator().validate_timestamp_and_nonce(
                'X', time.time(), '', self.request))

        self.assertFalse(
            BLTIRequestValidator().validate_timestamp_and_nonce(
                'X', time.time() - 65, '', self.request))

        self.assertFalse(
            BLTIRequestValidator().validate_timestamp_and_nonce(
                'X', time.time() + 65, '', self.request))

        self.assertFalse(
            BLTIRequestValidator().validate_timestamp_and_nonce(
                'X', '1234567890', '', self.request))


class CanvasRolesTest(TestCase):
    def setUp(self):
        self.params = LTI_LAUNCH_PARAMS
        self.launch_data = CanvasData(**self.params)

    def _authorize(self, role):
        return Roles(self.launch_data).authorize(role=role)

    def test_authorize_nodata(self):
        # no blti object
        self.assertRaises(
            ImproperlyConfigured, Roles(None).authorize, role='member')

    def test_authorize_public(self):
        self.assertEquals(None, self._authorize('public'))
        self.assertEquals(None, self._authorize(None))

    def test_authorize_member(self):
        self.params['roles'] = 'User'
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='member')

        self.params['roles'] = 'Administrator'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('member'))

        self.params['roles'] = 'urn:lti:instrole:ims/lis/Observer'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('member'))

    def test_authorize_admin(self):
        self.assertRaises(
            BLTIException, Roles(self.launch_data).authorize, role='admin')

        self.params['roles'] = 'Learner'
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='admin')

        self.params['roles'] = 'urn:lti:instrole:ims/lis/Administrator'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        self.params['roles'] = 'urn:lti:role:ims/lis/TeachingAssistant'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        self.params['roles'] = 'Learner,ContentDeveloper'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('admin'))

        # ignores ext_roles
        self.params['roles'] = 'Learner'
        self.params['ext_roles'] = 'urn:lti:role:ims/lis/Instructor'
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='admin')

    def test_authorize_specific(self):
        self.params['roles'] = 'Learner'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('Learner'))

        self.params['roles'] = 'urn:lti:role:ims/lis/Learner'
        self.launch_data = CanvasData(**self.params)
        self.assertEquals(None, self._authorize('Learner'))

        self.params['roles'] = 'Learner,ContentDeveloper'
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='Instructor')

        # unknown role
        self.params['roles'] = 'Learner'
        self.assertRaises(
            BLTIException, Roles(
                CanvasData(**self.params)).authorize, role='Manager')


class BLTI1p1SessionTest(TestCase):
    def setUp(self):
        self.request = RequestFactory().post(
            '/test', data=LTI_LAUNCH_PARAMS, secure=True)
        SessionMiddleware(get_response=mock.MagicMock()).process_request(
            self.request)

    def test_set_session(self):
        blti = BLTI()
        self.assertFalse(LTI_DATA_KEY in self.request.session)
        blti.set_session(self.request)
        self.assertTrue(LTI_DATA_KEY in self.request.session)

    def test_get_session(self):
        blti = BLTI()
        self.assertRaises(BLTIException, blti.get_session, self.request)

        blti.set_session(self.request)
        self.assertEqual(blti.get_session(self.request), {})

    def test_filter_oauth_params(self):
        data = LTI_LAUNCH_PARAMS

        self.assertEquals(len(data), 43)
        self.assertEquals(data['oauth_consumer_key'], 'XXXXXXXXXXXXXX')

        blti = BLTI()
        blti.set_session(self.request, **data)
        blti_data = blti.get_session(self.request)
        self.assertEquals(len(blti_data), 36)
        self.assertRaises(KeyError, lambda: blti_data['oauth_consumer_key'])


class BLTIDecoratorTest(TestCase):
    def setUp(self):
        self.request = RequestFactory().post(
            '/test', data=LTI_LAUNCH_PARAMS, secure=True)
        SessionMiddleware(get_response=mock.MagicMock()).process_request(
            self.request)

    def test_log_response_time(self):
        func = mock.Mock()
        func.__name__ = 'test'
        decorated_func = log_response_time(func)

        with self.assertLogs(level='INFO') as cm:
            response = decorated_func(self.request, 'test1', test_id=123)

        self.assertIn(
            'INFO:blti.performance:user: None, method: '
            'django.core.handlers.wsgi.test, args: (\'test1\',), '
            'kwargs: {\'test_id\': 123}, time: ',
            cm.output[0])



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from .base import BLTIView
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from blti.config import get_tool_conf, get_launch_data_storage
from blti.exceptions import BLTIException
from blti.validators import BLTIRequestValidator
from pylti1p3.exception import OIDCException
from pylti1p3.contrib.django import DjangoMessageLaunch
from oauthlib.oauth1.rfc5849.endpoints.signature_only import (
    SignatureOnlyEndpoint)
import logging


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class BLTILaunchView(BLTIView):
    http_method_names = ['get', 'post']

    def post(self, request, *args, **kwargs):
        return self._launch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return self._launch(request, *args, **kwargs)

    def _launch(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def dispatch(self, request, *args, **kwargs):
        try:
            launch_data = self.validate_1p1(request)
            logger.debug(f"LTI 1.1 launch")
        except BLTIException as ex:
            try:
                launch_data = self.validate_1p3(request)
                logger.debug(f"LTI 1.3 launch")
            except OIDCException as ex:
                logger.error(f"LTI authentication failure: {ex}")
                self.template_name = 'blti/401.html'
                return self.render_to_response(
                    {'LTI authentication failure': str(ex)}, status=401)
            except Exception as ex:
                logger.error(f"LTI launch error: {ex}")
                self.template_name = 'blti/401.html'
                return self.render_to_response(
                    {'LTI launch failure': str(ex)}, status=401)

        self.set_session(**launch_data)
        return super(BLTILaunchView, self).dispatch(request, *args, **kwargs)

    def validate_1p3(self, request):
        tool_conf = get_tool_conf()
        launch_data_storage = get_launch_data_storage()
        message_launch = DjangoMessageLaunch(
            request, tool_conf, launch_data_storage=launch_data_storage)
        message_launch_data = message_launch.get_launch_data()
        return message_launch_data

    def validate_1p1(self, request):
        request_validator = BLTIRequestValidator()
        endpoint = SignatureOnlyEndpoint(request_validator)
        uri = request.build_absolute_uri()
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        valid, oauth_req = endpoint.validate_request(
            uri, request.method, request.body, headers)

        # if non-ssl fixup scheme to validate signature generated
        # on the other side of ingress
        if not valid and uri.startswith('http:') and request.is_secure():
            valid, oauth_req = endpoint.validate_request(
                f"https{uri[4:]}", request.method, request.body, headers)

        if not valid:
            raise BLTIException('Invalid OAuth Request')

        blti_params = dict(oauth_req.params)
        return blti_params



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.urls import re_path
from blti.views import login, get_jwks, BLTIRawView


urlpatterns = [
    re_path(r'^login/?$', login, name='login'),
    re_path(r'^jwks/?$', get_jwks, name='jwks'),
    re_path(r'^$', BLTIRawView.as_view(), name='lti-launch-data'),
]

if getattr(settings, 'LTI_DEVELOP_APP', False):
    from blti.views.develop import BLTIDevPrepare, BLTIDevLaunch

    urlpatterns += [
        re_path(r'^dev/?$', BLTIDevPrepare.as_view(), name='dev-prepare'),
        re_path(r'^dev/launch/$', BLTIDevLaunch.as_view(), name='dev-launch'),
    ]



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


import json
from .launch import BLTILaunchView


class BLTIRawView(BLTILaunchView):
    template_name = 'blti/raw.html'
    authorized_role = 'admin'

    def get_context_data(self, **kwargs):
        params = [(k, getattr(self.blti, k)) for k in sorted(
            vars(self.blti)) if not k.startswith("_")]

        for role in ['is_member', 'is_administrator', 'is_staff',
                     'is_instructor', 'is_teaching_assistant',
                     'is_student', 'is_designer']:
            params.append((role, getattr(self.blti, role)))

        return {
            'digested_lti_params': params,
            'raw_lti_params': json.dumps(self.get_session(), indent=4)
        }



#!/usr/bin/env python

# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import django
from django.test.utils import get_runner
from django.conf import settings
import sys
import os


def load_settings():
    os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
    django.setup()


def run_tests(*test_args):
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(test_args)
    if failures:
        sys.exit(bool(failures))


if __name__ == "__main__":
    load_settings()
    run_tests(*sys.argv[1:])



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.db import models
from pylti1p3.roles import (
    AbstractRole, StaffRole, TeacherRole, TeachingAssistantRole,
    DesignerRole, ObserverRole, TransientRole)
import re


LTI_DATA_CLAIM_BASE = 'https://purl.imsglobal.org/spec/lti/claim/'


class LTILaunchData(object):
    def __init__(self, **data):
        self._data = data
        self.platform_name = self._data.get(
            'tool_consumer_info_product_family_code',
            self.claim_tool_platform('product_family_code'))

        self.is_member = self._1p1_roles(
            ['member']) or (
                StaffRole(self._data).check() or
                TeacherRole(self._data).check() or
                TeachingAssistantRole(self._data).check() or
                StudentRole(self._data).check() or
                ObserverRole(self._data).check())

        self.is_administrator = self._1p1_roles(
            ['admin']) or (
                TeacherRole(self._data).check() or
                TeachingAssistantRole(self._data).check() or
                DesignerRole(self._data).check())

        self.is_staff = self._1p1_roles(
            ['Administrator']) or (
                StaffRole(self._data).check())

        self.is_instructor = self._1p1_roles(
            ['Instructor']) or (
                TeacherRole(self._data).check())

        self.is_teaching_assistant = self._1p1_roles(
            ['TeachingAssistant']) or (
                TeachingAssistantRole(self._data).check())

        self.is_student = self._1p1_roles(
            ['Learner']) or (
                StudentRole(self._data).check())

        self.is_designer = self._1p1_roles(
            ['ContentDeveloper']) or (
                DesignerRole(self._data).check())

    def _1p1_roles(self, valid_roles):
        RE_ROLE_NS = re.compile(
            r'^urn:lti:(?:inst|sys)?role:ims/lis/([A-Za-z]+)$')
        AGGREGATE_ROLES = {
            'member': ['Administrator', 'Instructor', 'TeachingAssistant',
                       'ContentDeveloper', 'Learner', 'Observer'],
            'admin': ['Administrator', 'Instructor', 'TeachingAssistant',
                      'ContentDeveloper'],
        }

        try:
            roles = self._data['roles'].split(',')
        except KeyError:
            return False

        for valid_role in valid_roles:
            if valid_role in AGGREGATE_ROLES:
                return self._1p1_roles(AGGREGATE_ROLES[valid_role])

            for role in roles:
                if role == valid_role:
                    return True

                m = RE_ROLE_NS.match(role)
                if m and m.group(1) and m.group(1) == valid_role:
                    return True

        return False

    def claim_tool_platform(self, key, default=None):
        return self._claim_data('tool_platform', key, default)

    def claim_lis(self, key, default=None):
        return self._claim_data('lis', key, default)

    def claim_context(self, key, default=None):
        return self._claim_data('context', key, default)

    def claim_resource_link(self, key, default=None):
        return self._claim_data('resource_link', key, default)

    def claim_launch_presentation(self, key, default=None):
        return self._claim_data('launch_presentation', key, default)

    def claim_custom(self, key, default=None):
        return self._claim_data('custom', key, default)

    def _claim_data(self, claim, key, default=None):
        # sniff at the 1.3 claim, then sniff at corresponding 1.1 key
        return self._data.get(self._claim(claim), {}).get(
            key, self._data.get(f"{claim}_{key}", default))

    def _claim(self, claim):
        # specific claim key
        return f"{LTI_DATA_CLAIM_BASE}{claim}"


class StudentRole(AbstractRole):
    _common_roles = ("Learner", "Member")
    _institution_roles = ("Student", "Learner", "Member")
    _context_roles = ("Learner", "Member")



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


import time
from logging import getLogger
from blti import BLTI

logger = getLogger(__name__)


def log_response_time(func):
    def wrapper(*args, **kwargs):
        self = args[0]

        start = time.time()
        try:
            val = func(*args, **kwargs)
        except Exception:
            raise
        finally:
            module = self.__module__
            function = func.__name__

            is_view_class = True
            if module == "django.core.handlers.wsgi":
                is_view_class = False

            if is_view_class:
                arg_str = str(args[2:])
            else:
                arg_str = str(args[1:])

            kw_str = str(kwargs)
            end = time.time()

            try:
                request = args[1]
                login_id = BLTI().get_session(request).get(
                    'custom_canvas_user_login_id')
            except Exception as ex:
                login_id = None

            arg_str = arg_str.replace("#", "___")
            kw_str = kw_str.replace("#", "___")

            logger.info(
                'user: {}, method: {}.{}, args: {}, kwargs: {}, '
                'time: {}'.format(
                    login_id, module, function, arg_str, kw_str, end - start))
        return val
    return wrapper



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from oauthlib.oauth1.rfc5849.request_validator import RequestValidator
from oauthlib.oauth1.rfc5849.utils import UNICODE_ASCII_CHARACTER_SET
from blti.exceptions import BLTIException
import time
import logging


logger = logging.getLogger(__name__)


class BLTIRequestValidator(RequestValidator):
    @property
    def allowed_signature_methods(self):
        return ['HMAC-SHA1']

    @property
    def dummy_client(self):
        return 'dummy'

    @property
    def client_key_length(self):
        return 12, 30

    @property
    def nonce_length(self):
        return 20, 50

    @property
    def enforce_ssl(self):
        return getattr(
            settings, 'LTI_ENFORCE_SSL', True)

    @property
    def safe_characters(self):
        return set(UNICODE_ASCII_CHARACTER_SET) | set('-_')

    def validate_client_key(self, client_key, request):
        return self.get_client_secret(
            client_key, request) != self.dummy_client

    def get_client_secret(self, client_key, request):
        try:
            return getattr(
                settings, 'LTI_CONSUMERS', {})[client_key]
        except KeyError:
            return self.dummy_client

    def validate_timestamp_and_nonce(self, client_key, timestamp, nonce,
                                     request, request_token=None,
                                     access_token=None):
        now = int(time.time())
        return (now - 60) <= int(timestamp) <= (now + 60)


class Roles(object):
    def __init__(self, launch_data):
        self.blti = launch_data

    def authorize(self, role='member'):
        if not hasattr(self, 'blti') or self.blti is None:
            raise ImproperlyConfigured(
                'Roles class requires a blti model')

        role = role.lower() if role else None
        if not role or role == 'public' or (
                role == 'member' and
                self.blti.is_member) or (
                    role == 'admin' and
                    self.blti.is_administrator) or (
                        role in ['administrator', 'sysadmin'] and
                        self.blti.is_staff) or (
                            role in ['instructor', 'teacher']
                            and self.blti.is_instructor) or (
                                role in ['teachingassistant', 'ta'] and
                                self.blti.is_teaching_assistant) or (
                                    role in ['student', 'learner'] and
                                    self.blti.is_student) or (
                                        role in [
                                            'contentdeveloper', 'designer'] and
                                        self.blti.is_designer):
            return

        raise BLTIException('You are not authorized to view this content')



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


import os
from django.conf import settings
from importlib import resources
from pylti1p3.tool_config import ToolConfJsonFile
from pylti1p3.contrib.django import DjangoCacheDataStorage


LTI1P3_CONFIG_DIRECTORY_NAME = 'lti_config'
LTI1P3_CONFIG_FILE_NAME = 'tool.json'


def get_tool_conf():
    return ToolConfJsonFile(get_lti_config_path())


def get_launch_data_storage():
    return DjangoCacheDataStorage()


def get_lti_config_directory():
    return os.environ.get(
        'LTI_CONFIG_DIRECTORY',
        os.path.join(settings.BASE_DIR, LTI1P3_CONFIG_DIRECTORY_NAME))


def get_lti_config_path():
    return os.path.join(get_lti_config_directory(), LTI1P3_CONFIG_FILE_NAME)


def get_lti_public_key_path(key_name):
    return os.path.join(get_lti_config_directory(), key_name)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.views.decorators.csrf import csrf_exempt
from blti.config import get_tool_conf


@csrf_exempt
def get_jwks(request):
    tool_conf = get_tool_conf()
    return JsonResponse(tool_conf.get_jwks(), safe=False)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

SECRET_KEY = 'fake_key'
LTI_DEVELOP_APP = 'blti'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'blti'
)
ROOT_URLCONF = 'blti.tests.urls'
MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.PersistentRemoteUserMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'blti.middleware.SessionHeaderMiddleware',
    'blti.middleware.LTISessionAuthenticationMiddleware',
    'blti.middleware.CSRFHeaderMiddleware',
    'blti.middleware.SameSiteMiddleware'
]
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)
STATIC_ROOT = '/static/'
STATIC_URL = '/static/'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from blti.exceptions import BLTIException
import json


LTI_DATA_KEY = 'lti_launch_data'


class BLTI(object):
    def set_session(self, request, **kwargs):
        if not request.session.exists(request.session.session_key):
            request.session.create()

        # filter oauth_* parameters
        for key in list(filter(
                lambda key: key.startswith('oauth_'), kwargs.keys())):
            kwargs.pop(key)

        request.session[LTI_DATA_KEY] = json.dumps(kwargs)

    def get_session(self, request):
        try:
            return json.loads(request.session[LTI_DATA_KEY])
        except KeyError:
            raise BLTIException('Invalid Session')



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from blti.mock_data import Mock1p3Data
from blti.roles import roles_from_role_name
from django.views.generic import TemplateView
from django.urls import reverse, resolve
from django.urls.exceptions import NoReverseMatch
from importlib import import_module
from blti import BLTIException


class BLTIDevPrepare(TemplateView):
    template_name = 'blti/develop/prepare.html'


class BLTIDevLaunch(TemplateView):
    template_name = 'blti/develop/launch.html'

    def get_context_data(self, **kwargs):
        # strategy here is to reach behind the curtain, overriding
        # the 1.3 launch auth method with call that return mock data
        role = self.request.GET.get('role', '')
        campus = self.request.GET.get('campus', '')
        mock_jwt = Mock1p3Data().launch_data()

        # insert campus
        mock_jwt["https://purl.imsglobal.org/spec/lti/claim/custom"][
            "canvas_account_sis_id"] = (
                f"uwcourse:{campus}:a-and-s:pych:psych")

        # insert role
        claim, roles = roles_from_role_name(['User', role])
        mock_jwt[claim] = roles

        launch_module, launch_class = self._launch_module_and_class()
        launch_class.validate_1p3 = lambda launch_class, request: mock_jwt

        return {
            'role': role,
            'campus': campus
        }

    def _launch_module_and_class(self):
        resolver_match = resolve(self._lti_app())
        components = resolver_match._func_path.split('.')
        launch_module = '.'.join(components[:-1])
        launch_class = components[-1]
        module = import_module(launch_module)

        return module, getattr(module, launch_class)

    def _lti_app(self):
        try:
            return reverse('lti-launch')
        except NoReverseMatch:
            raise BLTIException("Cannot find 'lti-launch' url")



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


"""
Cross Site Request Forgery and Session HTTP Header Middleware.

This module provides middleware that implements protection
against request forgeries from other sites.
"""

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin
from django.contrib.auth import authenticate, login
from blti import BLTI
from blti.exceptions import BLTIException


class CSRFHeaderMiddleware(MiddlewareMixin):
    def process_request(self, request):
        csrf_token = request.META.get('HTTP_X_CSRFTOKEN', None)
        if csrf_token is not None:
            csrf_token_name = settings.CSRF_COOKIE_NAME
            request.COOKIES[csrf_token_name] = csrf_token


class SessionHeaderMiddleware(MiddlewareMixin):
    def process_request(self, request):
        session_id = request.META.get('HTTP_X_SESSIONID', None)
        if session_id is not None:
            session_key = settings.SESSION_COOKIE_NAME
            request.COOKIES[session_key] = session_id


class LTISessionAuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            lti_launch_parameters = BLTI().get_session(request)
            lti_user = lti_launch_parameters.get(
                "https://purl.imsglobal.org/spec/lti/claim/custom", {}).get(
                    "canvas_user_login_id")
            if lti_user:
                user = authenticate(request, remote_user=lti_user)
                login(request, user)
        except BLTIException:
            pass

        return self.get_response(request)


class SameSiteMiddleware(MiddlewareMixin):
    def process_response(self, request, response):
        if 'sessionid' in response.cookies:
            response.cookies['sessionid']['samesite'] = 'None'
        if 'csrftoken' in response.cookies:
            response.cookies['csrftoken']['samesite'] = 'None'
        return response



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from .login import login
from .jwks import get_jwks
from .base import BLTIView
from .launch import BLTILaunchView
from .raw import BLTIRawView
from .rest_dispatch import RESTDispatch



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.conf.urls import include
from django.urls import re_path
from blti.views import BLTIRawView


urlpatterns = [
    re_path(r'^$', BLTIRawView.as_view(), name='lti-launch'),
    re_path(r'^blti/', include('blti.urls')),
]



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.views.generic.base import TemplateView
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from blti import BLTI
from blti.models import CanvasData
from blti.exceptions import BLTIException
from blti.validators import Roles


@method_decorator(csrf_exempt, name='dispatch')
class BLTIView(TemplateView):
    authorized_role = 'member'

    def dispatch(self, request, *args, **kwargs):
        try:
            self.validate(request)
        except BLTIException as err:
            self.template_name = 'blti/401.html'
            return self.render_to_response({'error': str(err)}, status=401)

        return super(BLTIView, self).dispatch(request, *args, **kwargs)

    def render_to_response(self, context, **kwargs):
        response = super(BLTIView, self).render_to_response(context, **kwargs)
        self.add_headers(response=response, **kwargs)
        return response

    def add_headers(self, **kwargs):
        pass

    def set_session(self, **kwargs):
        BLTI().set_session(self.request, **kwargs)

    def get_session(self):
        return BLTI().get_session(self.request)

    def validate(self, request):
        self.blti = self.launch_data_model()

        if request.method != 'OPTIONS':
            self.authorize(self.authorized_role)

    def authorize(self, role):
        Roles(self.blti).authorize(role=role)

    def launch_data_model(self):
        return CanvasData(**self.get_session())



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.conf import settings
from importlib import resources
import json
import os
import logging


MOCK_JWT_FILE = ('resources', 'lti1p3', 'file', 'jwt.json')
logger = logging.getLogger(__name__)


class Mock1p3Data:
    def launch_data(self):
        mock_jwt_file = os.path.join(*MOCK_JWT_FILE)
        mock_jwt_path = os.path.join(
            resources.files('blti'), mock_jwt_file)

        for app in settings.INSTALLED_APPS:
            try:
                path = os.path.join(resources.files(app), mock_jwt_file)
            except Exception:
                continue

            logger.debug(f"Checking for mock jwt: {path}")
            if path != mock_jwt_path and os.path.exists(path):
                mock_jwt_path = path
                break

        logger.debug(f"Using mock jwt: {mock_jwt_path}")
        with open(mock_jwt_path, 'r') as f:
            return json.load(f)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from django.conf import settings
from django.test import TestCase, Client
from django.urls import reverse


class TestLaunchViews(TestCase):
    def setup(self, request, *args, **kwargs):
        self.client = Client()

    def test_dev_launch_view(self):
        response = self.client.get(reverse('dev-launch'), {
            'role': 'Instructor',
            'campus': 'seattle',
        })

        self.assertEqual(response.status_code, 200)

        response = self.client.post(reverse('lti-launch'))
        self.assertEqual(response.status_code, 200)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.core.management.base import BaseCommand, CommandError
from jwcrypto.jwk import JWK
from Crypto.PublicKey import RSA
import json


KEY_LENGTH = 4096


class Command(BaseCommand):
    help = 'Generate RSA keys and JWK for JWT signing'

    def add_arguments(self, parser):
        parser.add_argument('private_key_file', type=str, nargs=1)
        parser.add_argument('public_key_file', type=str, nargs=1)
        parser.add_argument('jwt_file', type=str, nargs=1)

    def create_keys(self):
        private_key = RSA.generate(KEY_LENGTH)
        return (private_key.exportKey(format='PEM'),
                private_key.publickey().exportKey(format='PEM'))

    def create_jwk(self, public_key):
        jwk_obj = JWK.from_pem(public_key)
        public_jwk = json.loads(jwk_obj.export_public())
        public_jwk['alg'] = 'RS256'
        public_jwk['use'] = 'sig'
        return json.dumps(public_jwk)

    def handle(self, *args, **options):
        private_key, public_key = self.create_keys()
        public_jwk = self.create_jwk(public_key)

        with open(options['private_key_file'][0], 'w') as f:
            f.write(private_key.decode('utf-8'))

        with open(options['public_key_file'][0], 'w') as f:
            f.write(public_key.decode('utf-8'))

        with open(options['jwt_file'][0], 'w') as f:
            f.write(public_jwk)



# Copyright 2024 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


from .base import LTILaunchData


class CanvasData(LTILaunchData):
    def __init__(self, **data):
        super(CanvasData, self).__init__(**data)

        if self.platform_name != 'canvas':
            raise ValueError('This is not a Canvas LTI launch')

        # Canvas internal IDs
        self.canvas_course_id = self.claim_custom('canvas_course_id')
        self.canvas_user_id = self.claim_custom('canvas_user_id')
        self.canvas_account_id = self.claim_custom('canvas_account_id')

        # SIS IDs
        self.course_sis_id = self.claim_lis('course_offering_sourcedid')
        self.user_sis_id = self.claim_lis('person_sourcedid')
        self.account_sis_id = self.claim_custom('canvas_account_sis_id')

        # Course attributes
        self.course_short_name = self.claim_context('label')
        self.course_long_name = self.claim_context('title')

        # User attributes
        self.user_login_id = self.claim_custom('canvas_user_login_id')
        self.user_full_name = data.get(
            'lis_person_name_full', data.get('name'))
        self.user_first_name = data.get(
            'lis_person_name_given', data.get('given_name'))
        self.user_last_name = data.get(
            'lis_person_name_family', data.get('family_name'))
        self.user_email = data.get(
            'lis_person_contact_email_primary', data.get('email'))
        self.user_avatar_url = data.get(
            'user_image', data.get('picture'))

        # LTI app attributes
        self.link_title = self.claim_resource_link('title')
        self.return_url = self.claim_launch_presentation('return_url')

        # Canvas hostname
        self.canvas_api_domain = self.claim_custom('canvas_api_domain')
